xquery version "3.1";
(:~
 : The TextGrid Client offers interfaces to the main features of TextGrid,
 : when they are available via REST.
 : @version 1.1
 :)

module namespace client="https://sade.textgrid.de/ns/client";

declare namespace http="http://expath.org/ns/http-client";
declare namespace rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#";
declare namespace sparql-results="http://www.w3.org/2005/sparql-results#";
declare namespace tgmd="http://textgrid.info/namespaces/metadata/core/2010";

(:~
 : @param $id - the TextGrid ID of the object, e.g. textgrid:1234.5
 :)

declare function client:get-metadata($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as node() {
    let $request-url := client:make-metadata-request-URL($id, $tgcrud-url, $sid)
    let $result := client:send-metadata-request($request-url, $id)
    return
        if( count($result) != 1 ) then
            <error> { $id } </error>
        else
            $result
};



declare function client:make-metadata-request-URL($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as xs:string {
    $tgcrud-url || "/" || $id || "/metadata?sessionId=" || $sid
};


declare function client:send-metadata-request($request-url as xs:string,
    $id as xs:string)
as node()? {
    let $result := client:make-request($request-url)
    return
        if ($result//tgmd:MetadataContainerType) then
            $result//tgmd:MetadataContainerType
        else
            let $error := 
            (: error(QName("https://sade.textgrid.de/ns/client", "client02"), :)
             ("No metadata object could be fetched for " 
            || $id || " from " || $request-url || ".")
            return util:log("error", $error)
};


declare function client:get-data-request-objects($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as item()+ {
    let $request-url := client:make-data-request-URL($id, $tgcrud-url, $sid)
    return
        client:make-request($request-url)
};


declare function client:get-data($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as item() {
    client:get-data-request-objects($id, $tgcrud-url, $sid)[2]
};

declare function client:get-data-request-header($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as item()+ {
    let $request-url := client:make-data-request-URL($id, $tgcrud-url, $sid)
    let $header :=
      <http:request method="head">
        <http:header name="Connection" value="close" />
      </http:request>
    return
        try {
            http:send-request($header, $request-url)
        } catch * {
            error(QName("https://sade.textgrid.de/ns/client", "client03"), "No header could be fetched from " || $request-url || ".")
        }
};


declare function client:make-data-request-URL($id as xs:string,
    $tgcrud-url as xs:string,
    $sid as xs:string?)
as xs:string {
    $tgcrud-url || "/" || $id || "/data?sessionId=" || $sid
};


declare function client:make-request($request-url as xs:string)
as item()+ {
    try {
        let $header :=
            <http:request method="get">
                <http:header name="Connection" value="close" />
            </http:request>
        return
            http:send-request($header, $request-url)
        } catch * {
            let $header :=
                <http:request method="get" override-media-type="text/plain">
                    <http:header name="Connection" value="close" />
                </http:request>
            let $response := http:send-request($header, $request-url)
            return
                ($response[1],
                <error time="{current-dateTime()}">
                            <code>{$err:code}</code>
                            <line>{$err:line-number}</line>
                            <column>{$err:column-number}</column>
                            <description>{$err:description}</description>
                            <value>{$err:value}</value>
                            <module>{$err:module}</module>
                            <additional>{$err:additional}</additional>
                            <document>{$response[2]}</document>
                </error>)
        }
};

(:~ Returns a list of the latest TextGrid items within a given aggregation.
 : @param tguri – the URI of any tg.aggregation
 : @param rdfstore – url to the triple store
 : @return sequence of strings (textgrid URIs) or empty sequence
 :)
declare function client:get-latest-aggregated-uris($tguri as xs:string,
    $rdfstore as xs:string)
as xs:string* {
    let $sparql-result := client:get-sparql-result($tguri, $rdfstore)
    let $uris := client:get-unrevisioned-aggregated-uris($sparql-result)
    for $uri in $uris
        let $max-rev := client:find-max-rev($sparql-result, $uri)
        return
            if (exists($max-rev)) then
                $uri || '.' || $max-rev
            else
                $uri
};


declare function client:get-sparql-result($tguri as xs:string,
    $rdfstore as xs:string)
as element(sparql-results:sparql) {
    let $query := client:make-sparql-query-for-aggregated-uris($tguri)
    return
        client:sparql($query, $rdfstore)
};


declare function client:make-sparql-query-for-aggregated-uris($tguri as xs:string)
as xs:string {
    concat("
    PREFIX ore:<http://www.openarchives.org/ore/terms/>
    PREFIX tg:<http://textgrid.info/relation-ns#>
    
    SELECT ?s
        WHERE { <",$tguri,"> (ore:aggregates/tg:isBaseUriOf|ore:aggregates)* ?s.
        FILTER not exists { ?s tg:isDeleted true } . }")
    => normalize-space()
};


declare function client:get-unrevisioned-aggregated-uris($sparql-result as element(sparql-results:sparql))
as xs:string+ {
    let $aggregated := $sparql-result//sparql-results:uri[not(. = "")]
    let $unrevisioned :=
        for $uri in $aggregated return
            if (contains($uri, ".")) then
                substring-before($uri,'.')
            else
                $uri
    return
        distinct-values($unrevisioned)
};


declare function client:find-max-rev($sparql-result as element(sparql-results:sparql),
    $current-uri as xs:string)
as xs:integer? {
    let $uris := $sparql-result//sparql-results:uri[starts-with(., $current-uri)]
    let $revisions := client:get-revisions-from-uris($uris)
    return
        max($revisions)
};

declare function client:get-revisions-from-uris($uris as xs:string+)
as xs:integer* {
    if (client:has-one-of-uris-revision($uris)) then
        for $uri in client:get-uris-with-revision($uris) return
            substring-after($uri, ".")
            => xs:integer()
    else
        ()
};

declare function client:has-one-of-uris-revision($uris as xs:string+)
as xs:boolean {
    exists($uris[contains(., ".")])
};

declare function client:get-uris-with-revision($uris as xs:string+)
as xs:string+ {
    $uris[contains(., ".")]
};


(:
 : Queries TextGrid RDF store. Mainly used to get URIs within a collection
 : @param $query - the SPARQL query as string
 : @param $rdfstore - URL of the public SPARQL endpoint
 : @return the result of the SPARQL query as XML node
 :)
declare function client:sparql($query as xs:string,
    $rdfstore as xs:string)
as node() {
    let $reqUrl := client:get-sparql-request-url($query, $rdfstore)
    let $header :=
      <http:request method="get">
        <http:header
          name = "accept"
          value = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
        />
        <http:header
          name = "connection"
          value = "close"
        />
      </http:request>

    return http:send-request($header, $reqUrl)[2]/node()
};



declare function client:get-sparql-request-url($query as xs:string,
    $rdfstore as xs:string)
as xs:string {
    let $urlEncodedQuery as xs:string := encode-for-uri($query)
    return
        string-join(($rdfstore, "?query=", $urlEncodedQuery))
};



declare function client:remove-prefix($tguri as xs:string) as xs:string {
    if (contains($tguri, ":")) then
        tokenize($tguri, ":")[position() gt 1] => string-join()
    else
        $tguri
};
